@@ -1,16 +1,21 @@
-stopifnot(require(RUnit, quietly=TRUE))
-stopifnot(require(Rcpp, quietly=TRUE))
-stopifnot(require(RcppParallel, quietly=TRUE))
+stopifnot(require(RUnit, quietly = TRUE))
+stopifnot(require(Rcpp, quietly = TRUE))
+stopifnot(require(RcppParallel, quietly = TRUE))
 
 ## Set a seed to make the test deterministic
 set.seed(42)
 
+# Don't use TBB
+Sys.setenv(RCPP_PARALLEL_BACKEND = "tinythread")
+
 ## Define tests
-suite <- defineTestSuite(name="RcppParallel Unit Tests",
-                         dirs=system.file("tests", package = "RcppParallel"))
+suite <- defineTestSuite(
+   name = "RcppParallel Unit Tests",
+   dirs = system.file("tests", package = "RcppParallel")
+)
 
 ## Based on practice in Rcpp to avoid some test failures
-Sys.setenv("R_TESTS"="")
+Sys.setenv("R_TESTS" = "")
 
 ## Run tests
 tests <- runTestSuite(suite)
@@ -22,9 +27,11 @@ printTextProtocol(tests)
 if (getErrors(tests)$nFail > 0) {
    stop("TEST FAILED!")
 }
+
 if (getErrors(tests)$nErr > 0) {
    stop("TEST HAD ERRORS!")
 }
+
 if (getErrors(tests)$nTestFunc < 1) {
    stop("NO TEST FUNCTIONS RUN!")
 }
